/*
 *  Get information about ULTRA driver
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#include "driver.h"
#include <stdarg.h>

#ifdef ULTRACFG_DEBUG_MEMORY
#define INFO_BUFFER_SIZE	( 700 * 1024 )
#else
#define INFO_BUFFER_SIZE	( 32 * 1024 )
#endif

int ultra_iprintf( ultra_info_buffer_t *buffer, char *fmt, ... )
{
  va_list args;
  int res;
  char sbuffer[ 512 ];
  
  if ( buffer -> stop ) return 0;
  va_start( args, fmt );
  res = vsprintf( sbuffer, fmt, args );
  va_end( args );
  if ( buffer -> size + res >= INFO_BUFFER_SIZE )
    {
      buffer -> stop = 1;
      return 0;
    }
  strcpy( buffer -> curr, sbuffer );
  buffer -> curr += res;
  buffer -> size += res;
  return res;
}

static void ultra_info_init( ultra_info_buffer_t *buffer )
{
#ifdef MODULE_PARM
  static char *kernel_version = UTS_RELEASE;
#else
  extern char kernel_version[];
#endif

  buffer -> curr = buffer -> buffer;
  *buffer -> buffer = 0;
  ultra_iprintf( buffer,
                       "Ultra Driver " VERSION ".\n"
  		       COPYRIGHT "\n"
                       "Compiled in " __DATE__ " for kernel %s"
#ifdef __SMP__
		       " (SMP)"
#endif
#ifdef MODVERSIONS
                       " with versioned symbols"
#endif
                       ".\n", kernel_version );

  ultra_memory_info( buffer );
  ultra_minor_info( buffer );
  ultra_info_card( buffer );
#ifdef ULTRACFG_DEBUG_MEMORY
  ultra_memory_debug( buffer );
#endif
}

static long ultra_info_read( struct file *file, char *buffer, long count )
{
  ultra_info_buffer_t *buf;
  int size, size1;
  
  buf = (ultra_info_buffer_t *)file -> private_data;
  if ( !buf ) return -EIO;
  if ( !buf -> size )
    ultra_info_init( buf );
  if ( file -> f_pos >= buf -> size ) return 0;
  size = buf -> size < count ? buf -> size : count;
  size1 = buf -> size - file -> f_pos;
  if ( size1 < size ) size = size1;
  MEMCPY_TOFS( buffer, buf -> buffer + file -> f_pos, size );
  file -> f_pos += size;
  return size;
}

static int ultra_info_open( unsigned short minor, struct file *file )
{
  ultra_info_buffer_t *buf;
  
  buf = (ultra_info_buffer_t *)ultra_malloc( sizeof( ultra_info_buffer_t ) );
  if ( !buf ) return -ENOMEM;
  MEMSET( buf, 0, sizeof( ultra_info_buffer_t ) ); 
  buf -> buffer = (char *)VMALLOC( INFO_BUFFER_SIZE );
  if ( !buf -> buffer ) {
    ultra_free( buf, sizeof( ultra_info_buffer_t ) );
    return -ENOMEM;
  }
  file -> private_data = buf;
  MOD_INC_USE_COUNT;
  return 0;
}

static int ultra_info_release( unsigned short minor, struct file *file )
{
  ultra_info_buffer_t *buf;
  
  buf = (ultra_info_buffer_t *)file -> private_data;
  file -> private_data = NULL;
  if ( buf ) {
    if ( buf -> buffer ) VFREE( buf -> buffer );
    ultra_free( buf, sizeof( ultra_info_buffer_t ) );
  }  
  MOD_DEC_USE_COUNT;
  return 0;
}

static long ultra_sndstat_read( struct file *file, char *buffer, long count )
{
  char *s = SNDSTAT_STRING;
  int size;

  size = strlen( s ) + 1;
  if ( file -> f_pos >= size || count < size - file -> f_pos )
    return 0;
  size -= file -> f_pos;
  MEMCPY_TOFS( buffer, s + file -> f_pos, size );
  file -> f_pos += size;
  return size;
}

static int ultra_sndstat_open( unsigned short minor, struct file *file )
{
  MOD_INC_USE_COUNT;
  return 0;
}

static int ultra_sndstat_release( unsigned short minor, struct file *file )
{
  MOD_DEC_USE_COUNT;
  return 0;
}

/*
 *
 */

static int ultra_info_unregister( void );

static struct ultra_stru_register_minor ultra_info_reg = {
  "ultrainfo",

  0,					/* don't verify soundcard # */

  { ULTRA_MINOR_UINFO, -1 },
  { -1 },
  
  ultra_info_unregister,		/* unregister */

  NULL,					/* lseek */
  ultra_info_read,			/* read */
  NULL,					/* write */
  ultra_info_open,			/* open */
  ultra_info_release,			/* release */
  NULL,					/* poll or select */
  NULL,					/* ioctl */
  NULL					/* mmap */
};

static int ultra_sndstat_unregister( void );

static struct ultra_stru_register_minor ultra_sndstat_reg = {
  "sndstat (OSS compatibility)",

  0,					/* don't verify soundcard # */

  { -1 },
  { ULTRA_MINOR_SNDSTAT, -1 },

  ultra_sndstat_unregister,		/* unregister */

  NULL,					/* lseek */
  ultra_sndstat_read,			/* read */
  NULL,					/* write */
  ultra_sndstat_open,			/* open */
  ultra_sndstat_release,		/* release */
  NULL,					/* poll or select */
  NULL,					/* ioctl */
  NULL					/* mmap */
};

static int ultra_info_unregister( void )
{
  ultra_unregister_minor( &ultra_info_reg );
  return 0;
}

static int ultra_sndstat_unregister( void )
{
  ultra_unregister_minor( &ultra_sndstat_reg );
  return 0;
}

int ultra_info_register( void )
{
  int err;

  if ( ( err = ultra_register_minor( &ultra_info_reg ) ) < 0 )
    return err;
  if ( ( err = ultra_register_minor( &ultra_sndstat_reg ) ) < 0 ) {
    ultra_unregister_minor( &ultra_info_reg );
    return err;
  }
  return 0;
}
