/*
 *  Initialization & resource registration routines
 *  Copyright (c) by Jaroslav Kysela <perex@jcu.cz>
 */

#include "driver.h"

int ultra_cards_count;
unsigned int ultra_cards_bitmap;
ultra_card_t *ultra_cards[ ULTRA_CARDS ];
ultra_card_t *ultra_cards_irq_index[ 16 ];

void ultra_driver_init( void )
{
  ultra_cards_count = 0;
  ultra_cards_bitmap = 0;
  MEMSET( &ultra_cards, 0, sizeof( ultra_cards ) );
  MEMSET( &ultra_cards_irq_index, 0, sizeof( ultra_cards_irq_index ) );
}

ultra_card_t *ultra_card_new( char *xid, char *name, void (*use_inc)( void ), void (*use_dec)( void ) )
{
  int idx;
  ultra_card_t *card;
  
  if ( !use_inc || !use_dec ) return NULL;
  card = (ultra_card_t *)ultra_malloc( sizeof( ultra_card_t ) );
  if ( !card ) return NULL;
  MEMSET( card, 0, sizeof( ultra_card_t ) );
  if ( xid )
    strncpy( card -> id, xid, 8 );
  if ( name )
    strncpy( card -> name, name, sizeof( card -> name ) - 1 );
  for ( idx = 0; idx < ULTRA_CARDS; idx++ ) {
    if ( ultra_cards_bitmap & (1 << idx) ) continue;
    card -> number = idx;
    if ( !card -> id[ 0 ] )
      sprintf( card -> id, "card%i", card -> number + 1 );
    card -> use_inc = use_inc;
    card -> use_dec = use_dec;
    return card;
  }
  ultra_free( card, sizeof( ultra_card_t ) );
  return NULL;			/* Oops.. Max number of soundcards reached!!! */
}

int ultra_card_free( ultra_card_t *card )
{
  ultra_unregister_interrupts( card );
  ultra_unregister_dma_channels( card );
  ultra_unregister_ioports( card );
  if ( card -> private_free )
    card -> private_free( card -> private );
  ultra_free( card, sizeof( ultra_card_t ) );
  return 0;
}

int ultra_card_register( ultra_card_t *card )
{
  unsigned long flags;
  
  CLI( &flags );
  ultra_cards_bitmap |= 1 << card -> number;
  ultra_cards[ ultra_cards_count++ ] = card;
  STI( &flags );
  return 0;
}

int ultra_card_unregister( ultra_card_t *card )
{
  unsigned long flags;
  
  if ( !card ) return -EINVAL;
  CLI( &flags );
  ultra_cards_bitmap &= ~(1 << card -> number);
  ultra_cards[ card -> number ] = NULL;
  ultra_cards_count--;
  STI( &flags );
  ultra_card_free( card );
  return 0;
}

char *ultra_get_id( ultra_card_t *ultra, char *id )
{
  strncpy( id, ultra -> id, 8 );
  id[ 8 ] = 0;
  return id;
}

void ultra_info_card( ultra_info_buffer_t *buffer )
{
  unsigned long flags;
  int idx, count;
  ultra_card_t *card;
  char id[ 9 ];
  
  ultra_iprintf( buffer, "\nRegistered soundcards:\n" );
  for ( idx = count = 0; idx < ULTRA_CARDS; idx++ ) {
    CLI( &flags );
    if ( ultra_cards_bitmap & (1 << idx) ) {
      count++;
      card = ultra_cards[ idx ];
      ultra_iprintf( buffer, "  %02i [%-8s]: %s\n", idx, ultra_get_id( card, id ), card -> name );
    }
    STI( &flags );
  }
  if ( !count ) {
    ultra_iprintf( buffer, "  --- no soundcards registered ---\n" );
  }
}

/*
 *  Misc...
 */

void ultra_delay( int loops )
{
  int i;

  while( loops-- > 0 )
      for ( i = 0; i < 16; i++ ) INB( 0x80 );
}      

/*
 *  I/O port registration
 */

int ultra_register_ioport( ultra_card_t *card, int port, int size, char *name )
{
  int idx;
  char *xname;
  unsigned long flags;
  
  if ( !card ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_PORTS; idx++ ) {
    if ( card -> ports_map & (1 << idx) ) continue;
    xname = ultra_malloc_strdup( name, SP_KERNEL );
    CLI( &flags );
    if ( check_region( port, size ) ) {
      STI( &flags );
      ultra_free_str( xname );
      return -EBUSY;
    }
    request_region( port, size, xname );
    card -> ports_map |= 1 << idx;
    card -> ports[ idx ].port = port;
    card -> ports[ idx ].size = size;
    card -> ports[ idx ].name = xname;
    STI( &flags );
    return 0;
  }
  return -ENOMEM;
}

int ultra_unregister_ioports( ultra_card_t *card )
{
  int idx;
  
  if ( !card ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_PORTS; idx++ ) {
    if ( card -> ports_map & (1 << idx) ) {
      release_region( card -> ports[ idx ].port, card -> ports[ idx ].size );
      ultra_free_str( card -> ports[ idx ].name );
    }
  }
  card -> ports_map = 0;
  return 0;
}

/*
 *  DMA channel registration
 */

int ultra_register_dma_channel( ultra_card_t *card, int idx, char *name, int number, int rsize, int *possible_numbers )
{
  struct ULTRA_STRU_DMA *dmaptr;

  if ( !card ) return -EINVAL;
  if ( idx < 0 || idx >= ULTRA_DMAS ) return -EINVAL;
  if ( card -> dmas_map & (1 << idx) ) return -EINVAL;
  if ( rsize == ULTRA_AUTO ) rsize = 128;
  if ( rsize < 4 || rsize > 128 ) return -EINVAL;
  if ( number == ULTRA_AUTO ) {
    if ( !possible_numbers ) return -EINVAL;
    for ( ; ( number = *possible_numbers ) >= 0 ; possible_numbers++ ) {
      if ( request_dma( number, "ultra test" ) ) continue;
      free_dma( number );
      break;
    }
    if ( number < 0 ) return -ENOMEM;
  }
  if ( number < 4 && rsize > 64 ) rsize = 64;
  if ( number < 0 || number > 7 ) return -EINVAL;
  dmaptr = &card -> dmas[ idx ];
  MEMSET( dmaptr, 0, sizeof( *dmaptr ) );
  dmaptr -> dma = number;
  dmaptr -> rsize = dmaptr -> ursize = rsize * 1024;
  dmaptr -> name = ultra_malloc_strdup( name, SP_KERNEL );
  if ( request_dma( number, dmaptr -> name ) ) {
    PRINTK( "ultra: unable to grab DMA %i for %s\n", number, dmaptr -> name );
    ultra_free_str( dmaptr -> name );
    dmaptr -> name = NULL;
    return -EBUSY;
  }
  card -> dmas_map |= 1 << idx;
  return 0;
}

int ultra_unregister_dma_channels( ultra_card_t *card )
{
  int idx;
  struct ULTRA_STRU_DMA *dmaptr;

  if ( !card ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_DMAS; idx++ ) {
    if ( !(card -> dmas_map & (1 << idx)) ) continue;
    dmaptr = &card -> dmas[ idx ];
    disable_dma( dmaptr -> dma );
    free_dma( dmaptr -> dma );
    ultra_free_str( dmaptr -> name );
  }
  card -> dmas_map = 0;
  return 0;
}

/*
 *  IRQ channel registration
 */

int ultra_register_interrupt( ultra_card_t *card, int number, ultra_irq_handler_t *handler, char *name, int *possible_numbers )
{
  int idx;
  char *xname;

  if ( !card ) return -EINVAL;
  if ( number == ULTRA_AUTO ) {
    for ( ; ( number = *possible_numbers ) >= 0; possible_numbers++ ) {
      if ( ultra_request_irq( number, handler, SA_ULTRA_FLAGS, "ultra test", card ) ) continue;
      ultra_free_irq( number, card );
      break;
    }
    if ( number < 0 ) return -ENOMEM;
  }
  if ( number < 0 || number > 15 ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_IRQS; idx++ ) {
    if ( card -> irqs_map & (1 << idx) ) continue;
    break;
  }
  if ( idx >= ULTRA_IRQS ) return -ENOMEM;
  xname = ultra_malloc_strdup( name, SP_KERNEL );
  if ( ultra_request_irq( number, handler, SA_ULTRA_FLAGS, xname, card ) ) {
    PRINTK( "ultra: unable to grab IRQ %i for %s\n", number, xname );
    ultra_free_str( xname );
    return -EBUSY;
  }
  ultra_cards_irq_index[ number ] = card;
  card -> irqs_map |= 1 << idx;
  card -> irqs[ idx ].number = number;
  card -> irqs[ idx ].name = xname;
  enable_irq( number );
  return 0;
}

int ultra_unregister_interrupts( ultra_card_t *card )
{
  int number, idx;
  
  if ( !card ) return -EINVAL;
  for ( idx = 0; idx < ULTRA_IRQS; idx++ ) {
    if ( card -> irqs_map & (1 << idx) ) {
      number = card -> irqs[ idx ].number;
      disable_irq( number );
      ultra_free_irq( number, card );
      ultra_free_str( card -> irqs[ idx ].name );
    }
  }
  card -> irqs_map = 0;
  return 0;
}
